# SPDX-License-Identifer: GPL-3.0-or-later

from abc import ABC, ABCMeta, abstractmethod
from typing import Callable, Optional

from gi.repository import Gio, GLib


class BaseMeta(ABCMeta, type(Gio.Application)):
    pass


class Cleanable(ABC):
    @abstractmethod
    def clean_resources(self) -> None:
        pass


class BaseGioApplication(Cleanable, Gio.Application, metaclass=BaseMeta):
    def get_dbus_connection2(self) -> Gio.DBusConnection:
        dbus_connection = super().get_dbus_connection()

        if not dbus_connection:
            raise Exception('No DBus connection available')

        return dbus_connection

    def stdin_add_watch(
        self,
        priority: int,
        condition: GLib.IOCondition,
        callback: Callable[[GLib.IOChannel, GLib.IOCondition, Optional[GLib.Variant]], None],
        user_data: Optional[GLib.Variant] = None
    ) -> None:
        pass


class ApplicationHandler(Cleanable):
    @abstractmethod
    def __init__(self, application: BaseGioApplication) -> None:
        pass



# SPDX-License-Identifer: GPL-3.0-or-later

import re
from typing import Any

from gi.repository import Gio, GLib

def get_variant(data: Any) -> GLib.Variant:
    if isinstance(data, str):
        return GLib.Variant.new_string(data)
    elif isinstance(data, int):
        return GLib.Variant.new_int32(data)
    elif isinstance(data, (list, tuple, set)):
        variant_builder: GLib.VariantBuilder = GLib.VariantBuilder.new(GLib.VariantType.new('av'))

        for value in data:
            variant_builder.add_value(GLib.Variant.new_variant(get_variant(value)))

        return variant_builder.end()
    elif isinstance(data, dict):
        variant_builder = GLib.VariantBuilder.new(GLib.VariantType.new('a{sv}'))

        for key in data:
            if data[key] is None:
                continue

            key_string = str(key)

            variant_builder.add_value(
                GLib.Variant.new_dict_entry(
                    get_variant(key_string), GLib.Variant.new_variant(get_variant(data[key]))
                )
            )

        return variant_builder.end()
    else:
        raise Exception(f"Unknown data type: {type(data)}")

# https://wiki.gnome.org/Projects/GnomeShell/Extensions/UUIDGuidelines
def is_uuid(uuid: str):
    return uuid is not None and re.match('[-a-zA-Z0-9@._]+$', uuid) is not None

def obtain_gio_settings(schema: str) -> Gio.Settings:
    source: Gio.SettingsSchemaSource = Gio.SettingsSchemaSource.get_default()

    if source is None:
        raise Exception("No Gio.Settings schemas are installed")

    if source.lookup(schema, True) is None:
        raise Exception("Settings schema with id `{schema}` is missing")

    return Gio.Settings.new(schema)



# SPDX-License-Identifer: GPL-3.0-or-later

from typing import Callable, Optional

from gi.repository import Gio, GLib, GObject

from .base import ApplicationHandler, BaseGioApplication
from .logs import get_logger


class Service(ApplicationHandler):
    def __init__(self, application: BaseGioApplication) -> None:
        if application.get_is_remote():
            return

        super().__init__(application)

        self._log = get_logger(self)
        self._application = application

        self.add_simple_action("create-notification", self.on_create_notification, 'a{sv}')
        self.add_simple_action("on-notification-clicked", self.on_notification_clicked, 's')
        self.add_simple_action("on-notification-action", self.on_notification_action, '(si)')

        GLib.timeout_add_seconds(5 * 60, self.on_service_timeout, None)

        self._application.hold()

        self._log.debug("Service started")

    def clean_resources(self) -> None:
        pass

    def add_simple_action(self, name: str, callback: Callable[..., None], parameter_type: str) -> None:
        action = Gio.SimpleAction.new(
            name,
            GLib.VariantType.new(parameter_type) if parameter_type is not None else None
        )
        action.connect('activate', callback)
        self._application.add_action(action)

    # Service events
    def on_create_notification(self, source: GObject.Object, request: GLib.Variant) -> None:
        self._log.debug('On create notification')

        request = request.unpack()

        notification = Gio.Notification.new(request['title'])
        notification.set_body(request['message'])
        notification.set_priority(Gio.NotificationPriority.NORMAL)
        notification.set_default_action_and_target(
            "app.on-notification-clicked",
            GLib.Variant.new_string(request['name'])
        )

        if 'buttons' in request:
            for button_id, button in enumerate(request['buttons']):
                notification.add_button_with_target(
                    button['title'],
                    "app.on-notification-action",
                    GLib.Variant.new_tuple(
                        GLib.Variant.new_string(request['name']),
                        GLib.Variant.new_int32(button_id)
                    )
                )

        self._application.send_notification(request['name'], notification)

    def on_notification_action(self, notification: GObject.Object, parameters: GLib.Variant) -> None:
        self._log.debug('Notification %s action: %s', *parameters.unpack())

        self._application.get_dbus_connection2().emit_signal(
            None,
            self._application.get_dbus_object_path(),
            self._application.get_application_id(),
            "NotificationAction",
            parameters
        )

    def on_notification_clicked(self, notification: GObject.Object, notification_name: GLib.Variant) -> None:
        self._log.debug('Notification %s clicked', notification_name)

        self._application.get_dbus_connection2().emit_signal(
            None,
            self._application.get_dbus_object_path(),
            self._application.get_application_id(),
            "NotificationClicked",
            GLib.Variant.new_tuple(notification_name)
        )

    def on_service_timeout(self, data: Optional[GObject.Object]) -> bool:
        self._log.debug('On service timeout')
        self._application.clean_resources()

        return False



# SPDX-License-Identifer: GPL-3.0-or-later

import logging
from typing import Any

def get_logger(obj: Any):
    return logging.getLogger(".".join((
        obj.__class__.__module__,
        obj.__class__.__qualname__))
    )


class NameAbbrFilter(logging.Filter):
    def filter(self, record: logging.LogRecord):
        modules = record.name.split(".")
        record.name_abbr = ".".join(
            [
                "_".join(
                    p[:1]
                    for p in m.split("_")
                )
                for m in modules[:-1]
            ] +
            [modules[-1]]
        )

        return True



# SPDX-License-Identifer: GPL-3.0-or-later

from __future__ import annotations

import signal
import sys
import traceback
from types import TracebackType
from typing import Any, Callable, Optional, Sequence

from gi.repository import GLib, Gio

from gnome_browser_connector.constants import CONNECTOR_ARG
from gnome_browser_connector.helpers import is_uuid

from .base import BaseGioApplication
from .logs import get_logger
from .connector import Connector
from .service import Service


class Application(BaseGioApplication):
    def __init__(self) -> None:
        Gio.Application.__init__(
            self,
            application_id='org.gnome.BrowserConnector',
            flags=Gio.ApplicationFlags.HANDLES_OPEN
        )

        self._log = get_logger(self)
        self._handler = None

        # Set custom exception hook
        sys.excepthook = self.default_exception_hook

        self.add_main_option(
            CONNECTOR_ARG,
            ord(CONNECTOR_ARG[:1]),
            GLib.OptionFlags.NONE,
            GLib.OptionArg.NONE,
            "Run as browser messaging host"
        )

        GLib.unix_signal_add(GLib.PRIORITY_DEFAULT, signal.SIGINT, self.on_sigint, None)

        self._stdin: GLib.IOChannel = GLib.IOChannel.unix_new(sys.stdin.fileno())
        self._stdin.set_encoding(None)
        self._stdin.set_buffered(False)

        self.stdin_add_watch(GLib.PRIORITY_DEFAULT, GLib.IOCondition.HUP, self.on_hup, None)
        self.stdin_add_watch(GLib.PRIORITY_DEFAULT, GLib.IOCondition.ERR, self.on_hup, None)

    def do_activate(self) -> None:
        pass

    def do_handle_local_options(self, options: GLib.VariantDict):
        is_service = (self.get_flags() & Gio.ApplicationFlags.IS_SERVICE).real
        if not is_service:
            self.set_flags(self.get_flags() | Gio.ApplicationFlags.IS_LAUNCHER)

        self.register()

        if is_service:
            self._handler = Service(self)
        elif options.contains(CONNECTOR_ARG):
            self._handler = Connector(self)

        return -1

    def do_open(self, files: Sequence[Gio.File], n_files: int, hint: str):
        for file in files:
            if file.get_uri_scheme() != 'gnome-extensions':
                continue

            uri: GLib.Uri = GLib.uri_parse(file.get_uri(), GLib.UriFlags.NON_DNS)
            uuid = uri.get_host()
            if not is_uuid(uuid):
                self._log.fatal(f"Wrong extension UUID passed: `{uuid}`")
                continue

            params = GLib.Uri.parse_params(uri.get_query(), -1, "&", GLib.UriParamsFlags.NONE)
            if 'action' in params:
                if params['action'] == 'install':
                    try:
                        Gio.DBusProxy.new_sync(
                            self.get_dbus_connection2(),
                            Gio.DBusProxyFlags.NONE,
                            None,
                            'org.gnome.Shell',
                            '/org/gnome/Shell',
                            'org.gnome.Shell.Extensions',
                            None
                        ).call_sync(
                            'InstallRemoteExtension',
                            GLib.Variant.new_tuple(GLib.Variant.new_string(uuid)),
                            Gio.DBusCallFlags.NONE,
                            -1,
                            None)
                    except GLib.GError as e:
                        self._log.fatal(f"Unable to install extension: {e.message}")
                        continue

    def stdin_add_watch(
        self,
        priority: int,
        condition: GLib.IOCondition,
        callback: Callable[[GLib.IOChannel, GLib.IOCondition, Optional[GLib.Variant]], None],
        user_data: Optional[GLib.Variant] = None
    ) -> None:
        GLib.io_add_watch(self._stdin, priority, condition, callback, None)

    # Is there any way to hook this to shutdown?
    def clean_resources(self) -> None:
        self._log.debug('Releasing resources')

        if self._handler:
            self._handler.clean_resources()

        self.release()

    def default_exception_hook(
        self,
        exception_type: type[BaseException],
        value: BaseException,
        tb: TracebackType
    ) -> None:
        self._log.fatal("Uncaught exception of type %s occured", exception_type)
        traceback.print_tb(tb)
        self._log.fatal("Exception: %s", value)

        self.clean_resources()

    # General events
    def on_hup(self, source: GLib.IOChannel, condition: GLib.IOCondition, data: Optional[GLib.Variant]):
        self._log.debug('On hup: %s', str(condition))
        self.clean_resources()

        return False

    def on_sigint(self, data: Optional[Any]):
        self._log.debug('On sigint')
        self.clean_resources()

        return False



# SPDX-License-Identifer: GPL-3.0-or-later

import logging
import os
import sys

from .application import Application
from .constants import CONNECTOR_ARG
from .logs import NameAbbrFilter

def main():
    logging.basicConfig(
        format="%(asctime)s: [%(process)d] %(levelname)s %(name_abbr)s %(message)s",
        level=getattr(
            logging,
            os.getenv("GNOME_BROWSER_CONNECTOR_LOGLEVEL", "warning").upper()
        ),
        stream=sys.stderr
    )
    logging.getLogger().handlers[0].addFilter(NameAbbrFilter())
    logging.debug('Main')

    app = Application()
    code = app.run(sys.argv)

    logging.debug('Quit')

    return code

def connector():
    ensure_argument_exists(CONNECTOR_ARG)
    return main()

def ensure_argument_exists(argument: str):
    if f"--{argument}" not in sys.argv:
        sys.argv.insert(1, f"--{argument}")



# SPDX-License-Identifer: GPL-3.0-or-later

from __future__ import annotations

import json
import struct
import sys
from typing import Any, Optional
from gi.repository import Gio, GLib, GObject

from .base import ApplicationHandler, BaseGioApplication
from .helpers import get_variant, is_uuid, obtain_gio_settings
from .logs import get_logger
from .version import __version__

SHELL_SCHEMA = "org.gnome.shell"


class Connector(ApplicationHandler):
    ENABLED_EXTENSIONS_KEY = "enabled-extensions"
    EXTENSION_DISABLE_VERSION_CHECK_KEY = "disable-extension-version-validation"
    DISABLE_USER_EXTENSIONS_KEY = "disable-user-extensions"

    # https://developer.chrome.com/extensions/nativeMessaging#native-messaging-host-protocol
    MESSAGE_LENGTH_SIZE = 4

    def __init__(self, application: BaseGioApplication) -> None:
        super().__init__(application)

        self._log = get_logger(self)
        self._application = application

        self._shell_appeared_id = 0
        self._shell_signal_id = 0
        self.disable_user_extensions_signal_id = None
        self.disable_version_check_signal_id = None

        self._shell_proxy = Gio.DBusProxy.new_sync(
            self._application.get_dbus_connection2(),
            Gio.DBusProxyFlags.NONE,
            None,
            'org.gnome.Shell',
            '/org/gnome/Shell',
            'org.gnome.Shell.Extensions',
            None
        )
        self._shell_settings = obtain_gio_settings(SHELL_SCHEMA)

        self._application.get_dbus_connection2().signal_subscribe(
            self._application.get_application_id(),
            self._application.get_application_id(),
            None,
            "/org/gnome/BrowserConnector",
            None,
            Gio.DBusSignalFlags.NONE,
            self.on_dbus_signal,
            None
        )

        self._application.stdin_add_watch(GLib.PRIORITY_DEFAULT, GLib.IOCondition.IN, self.on_input, None)

        self._application.hold()

        self._log.debug("Messaging host started")

    def clean_resources(self) -> None:
        self._log.debug('Releasing resources')

        if self._shell_appeared_id:
            Gio.bus_unwatch_name(self._shell_appeared_id)

        if self._shell_signal_id:
            self._application.get_dbus_connection2().signal_unsubscribe(self._shell_signal_id)

        if self.disable_user_extensions_signal_id:
            if self._shell_settings is not None:
                self._shell_settings.disconnect(self.disable_user_extensions_signal_id)

        if self.disable_version_check_signal_id:
            if self._shell_settings is not None:
                self._shell_settings.disconnect(self.disable_version_check_signal_id)

    def on_dbus_signal(
        self,
        connection: Gio.DBusConnection,
        sender_name: Optional[str],
        object_path: str,
        interface_name: str,
        signal_name: str,
        parameters: GLib.Variant,
        user_data: Optional[Any]
    ) -> None:
        self._log.debug('Signal %s from %s', signal_name, interface_name)

        if (
            interface_name == "org.gnome.Shell.Extensions" and
            signal_name == 'ExtensionStatusChanged'
        ):
            self.send_message({'signal': signal_name, 'parameters': parameters.unpack()})
        elif interface_name == self._application.get_application_id():
            if signal_name == 'NotificationAction':
                notification_name, button_id = parameters.unpack()

                self.send_message({
                    'signal': "NotificationAction",
                    'name': notification_name,
                    'button_id': button_id
                })
            elif signal_name == 'NotificationClicked':
                (notification_name,) = parameters.unpack()

                self.send_message({
                    'signal': "NotificationClicked",
                    'name': notification_name
                })

    def on_shell_appeared(
        self,
        connection: Gio.DBusConnection,
        name: str,
        name_owner: str
    ) -> None:
        self._log.debug('Signal: to %s', name)
        self.send_message({'signal': name})
        self._log.debug('Signal: from %s', name)

    def on_setting_changed(self, settings: Gio.Settings, key: str) -> None:
        if not key in (self.DISABLE_USER_EXTENSIONS_KEY, self.EXTENSION_DISABLE_VERSION_CHECK_KEY):
            return

        self._log.debug('on_setting_changed: %s=%s', key, settings.get_value(key).unpack())
        self.send_message({
            'signal': 'ShellSettingsChanged',
            'key': key,
            'value': settings.get_value(key).unpack()
        })

    # Native messaging events
    def on_input(
        self,
        source: GLib.IOChannel,
        condition: GLib.IOCondition,
        data: Optional[GObject.Object]
    ) -> Optional[bool]:
        self._log.debug('On input')
        text_length_bytes: bytes = source.read(self.MESSAGE_LENGTH_SIZE)

        if len(text_length_bytes) == 0:
            self._log.debug('Release condition: %s', condition)
            self._application.clean_resources()
            return

        # Unpack message length as 4 byte integer.
        text_length = struct.unpack(b'i', text_length_bytes)[0]

        # Read the text (JSON object) of the message.
        text: str = source.read(text_length).decode('utf-8')

        request = json.loads(text)

        if 'execute' in request:
            if 'uuid' in request and not is_uuid(request['uuid']):
                return

            self.process_request(request)

        return True

    def send_message(self, response: Any):
        """
        Helper function that sends a message to the webapp.
        :param response: dictionary of response data
        :return: None
        """

        message = json.dumps(response)
        message_length = len(message.encode('utf-8'))

        if message_length > 1024*1024:
            raise Exception(f'Too long message ({message_length}): "{message}"')

        try:
            stdout: GLib.IOChannel = GLib.IOChannel.unix_new(sys.stdout.fileno())
            stdout.set_encoding(None)
            stdout.set_buffered(False)

            stdout.write_chars(struct.pack(b'I', message_length), self.MESSAGE_LENGTH_SIZE)

            # Write the message itself.
            stdout.write_chars(message.encode('utf-8'), message_length)
        except IOError as e:
            raise Exception(f'IOError occured: {e.strerror}')

    def process_request(self, request: dict[str, Any]) -> None:
        self._log.debug("Execute: to %s", request['execute'])

        if request['execute'] == 'initialize':
            shell_version = self._shell_proxy.get_cached_property("ShellVersion")

            if shell_version is not None:
                if self.EXTENSION_DISABLE_VERSION_CHECK_KEY in self._shell_settings.keys():
                    disable_version_check: bool = self._shell_settings.get_boolean(
                        self.EXTENSION_DISABLE_VERSION_CHECK_KEY
                    )
                else:
                    disable_version_check = False

                if self.DISABLE_USER_EXTENSIONS_KEY in self._shell_settings.keys():
                    disable_user_extensions: bool = self._shell_settings.get_boolean(self.DISABLE_USER_EXTENSIONS_KEY)
                else:
                    disable_user_extensions = False

                supports = ['notifications', 'v6']

                self.send_message(
                    {
                        'success': True,
                        'properties': {
                            'connectorVersion': __version__,
                            'shellVersion': shell_version.unpack(),
                            'versionValidationEnabled': not disable_version_check,
                            'userExtensionsDisabled': disable_user_extensions,
                            'supports': supports
                        }
                    }
                )
            else:
                self.send_message(
                    {
                        'success': False,
                        'message': "no_gnome_shell"
                    }
                )

        elif request['execute'] == 'subscribeSignals':
            if not self._shell_appeared_id:
                self._shell_appeared_id: int = Gio.bus_watch_name_on_connection(
                    self._application.get_dbus_connection2(),
                    'org.gnome.Shell',
                    Gio.BusNameWatcherFlags.NONE,
                    self.on_shell_appeared,
                    None
                )

            if not self._shell_signal_id:
                self._shell_signal_id: int = self._application.get_dbus_connection2().signal_subscribe(
                    "org.gnome.Shell",
                    "org.gnome.Shell.Extensions",
                    "ExtensionStatusChanged",
                    "/org/gnome/Shell",
                    None,
                    Gio.DBusSignalFlags.NONE,
                    self.on_dbus_signal,
                    None
                )

            if not self.disable_user_extensions_signal_id:
                self.disable_user_extensions_signal_id = self._shell_settings.connect(
                    f"changed::{self.DISABLE_USER_EXTENSIONS_KEY}",
                    self.on_setting_changed)

            if not self.disable_version_check_signal_id:
                self.disable_version_check_signal_id = self._shell_settings.connect(
                    f"changed::{self.EXTENSION_DISABLE_VERSION_CHECK_KEY}",
                    self.on_setting_changed)

        elif request['execute'] == 'installExtension':
            self.dbus_call_response(
                "InstallRemoteExtension",
                GLib.Variant.new_tuple(GLib.Variant.new_string(request['uuid'])),
                "status"
            )

        elif request['execute'] == 'listExtensions':
            self.dbus_call_response("ListExtensions", None, "extensions")

        elif request['execute'] == 'enableExtension':
            uuids = self._shell_settings.get_strv(self.ENABLED_EXTENSIONS_KEY)

            extensions = []
            if 'extensions' in request:
                extensions = request['extensions']
            else:
                extensions.append({'uuid': request['uuid'], 'enable': request['enable']})

            for extension in extensions:
                if not is_uuid(extension['uuid']):
                    continue

                if extension['enable']:
                    if not extension['uuid'] in uuids:
                        uuids.append(extension['uuid'])
                elif extension['uuid'] in uuids:
                    uuids = [value for value in uuids if value != extension['uuid']]

            self._shell_settings.set_strv(self.ENABLED_EXTENSIONS_KEY, uuids)

            self.send_message({'success': True})

        elif request['execute'] == 'launchExtensionPrefs':
            self._shell_proxy.call("LaunchExtensionPrefs",
                                  GLib.Variant.new_tuple(GLib.Variant.new_string(request['uuid'])),
                                  Gio.DBusCallFlags.NONE,
                                  -1,
                                  None,
                                  None,
                                  None)

        elif request['execute'] == 'getExtensionErrors':
            self.dbus_call_response("GetExtensionErrors",
                                    GLib.Variant.new_tuple(GLib.Variant.new_string(request['uuid'])),
                                    "extensionErrors")

        elif request['execute'] == 'getExtensionInfo':
            self.dbus_call_response("GetExtensionInfo",
                                    GLib.Variant.new_tuple(GLib.Variant.new_string(request['uuid'])),
                                    "extensionInfo")

        elif request['execute'] == 'uninstallExtension':
            self.dbus_call_response("UninstallExtension",
                                    GLib.Variant.new_tuple(GLib.Variant.new_string(request['uuid'])),
                                    "status")

        elif request['execute'] == 'setUserExtensionsDisabled':
            self.send_message({
                'success': self.set_shell_boolean(
                    self.DISABLE_USER_EXTENSIONS_KEY,
                    request['disable']
                )
            })

        elif request['execute'] == 'setVersionValidationDisabled':
            self.send_message({
                'success': self.set_shell_boolean(
                    self.EXTENSION_DISABLE_VERSION_CHECK_KEY,
                    request['disable']
                )
            })

        elif request['execute'] == 'createNotification':
            Gio.DBusActionGroup.get(
                self._application.get_dbus_connection2(),
                self._application.get_application_id(),
                self._application.get_dbus_object_path()
            ).activate_action('create-notification', get_variant({
                'name': request['name'],
                'title': request['options']['title'],
                'message': request['options']['message'],
                'buttons': request['options']['buttons']
            }))

        elif request['execute'] == 'removeNotification':
            self._application.withdraw_notification(request['name'])

        self._log.debug('Execute: from %s',  request['execute'])

    # Helpers
    def dbus_call_response(self, method: str, parameters: Optional[GLib.Variant], result_property: str):
        try:
            result = self._shell_proxy.call_sync(method,
                                                parameters,
                                                Gio.DBusCallFlags.NONE,
                                                -1,
                                                None)

            self.send_message({'success': True, result_property: result.unpack()[0]})
        except GLib.GError as e:
            self.send_error(e.message)

    def send_error(self, message: str):
        self.send_message({'success': False, 'message': message})

    def set_shell_boolean(self, key: str, value: bool) -> bool:
        if key in self._shell_settings.keys():
            return self._shell_settings.set_boolean(key, True if value else False)

        return False
